### EXAMPLE 1: demonstrates how measurement error can lead to a type sign error in a covariate
### What kind of data invalidates fong + tyler?
### Even when you have a good predictor, if it's biased against a covariate you can get the wrong sign.
### Even when you include the proxy variable in the regression.
### But with some ground truth and multiple imputation, you can fix it.

library(argparser)
library(mecor)
library(ggplot2)
library(data.table)
library(filelock)
library(arrow)
library(Amelia)
library(Zelig)
library(predictionError)
options(amelia.parallel="no",
        amelia.ncpus=1)
setDTthreads(40)

source("simulation_base.R")

## SETUP:
### we want to estimate x -> y; x is MAR
### we have x -> k; k -> w; x -> w is used to predict x via the model w.
### A realistic scenario is that we have an NLP model predicting something like "racial harassment" in social media comments
### The labels x are binary, but the model provides a continuous predictor

### simulation:
#### how much power do we get from the model in the first place? (sweeping N and m)
#### 

## one way to do it is by adding correlation to x.obs and y that isn't in w.
## in other words, the model is missing an important feature of x.obs that's related to y.
simulate_data <- function(N, m, B0, Bxy, Bzy, seed, prediction_accuracy=0.73){
    set.seed(seed)

    # make w and y dependent
    z <- rbinom(N, 1, 0.5)
    x <- rbinom(N, 1, 0.5)

    ystar <- Bzy * z + Bxy * x + B0
    y <- rbinom(N,1,plogis(ystar))

    # glm(y ~ x + z, family="binomial")

    df <- data.table(x=x,y=y,ystar=ystar,z=z)

    if(m < N){
        df <- df[sample(nrow(df), m), y.obs := y]
    } else {
        df <- df[, y.obs := y]
    }
    
    odds.y1 <- qlogis(prediction_accuracy)
    odds.y0 <- qlogis(prediction_accuracy,lower.tail=F)

    df[y==0,w:=plogis(rlogis(.N,odds.y0))]
    df[y==1,w:=plogis(rlogis(.N,odds.y1))]

    df[,w_pred := as.integer(w > 0.5)]

    print(mean(df[x==0]$y == df[x==0]$w_pred))
    print(mean(df[x==1]$y == df[x==1]$w_pred))
    print(mean(df$w_pred == df$y))
    return(df)
}

parser <- arg_parser("Simulate data and fit corrected models")
parser <- add_argument(parser, "--N", default=1000, help="number of observations of w")
parser <- add_argument(parser, "--m", default=500, help="m the number of ground truth observations")
parser <- add_argument(parser, "--seed", default=17, help='seed for the rng')
parser <- add_argument(parser, "--outfile", help='output file', default='example_2.feather')
parser <- add_argument(parser, "--y_explained_variance", help='what proportion of the variance of y can be explained?', default=0.005)
parser <- add_argument(parser, "--prediction_accuracy", help='how accurate is the predictive model?', default=0.72)
## parser <- add_argument(parser, "--x_bias_y1", help='how is the classifier biased when y = 1?', default=-0.75)
## parser <- add_argument(parser, "--x_bias_y0", help='how is the classifier biased when y = 0 ?', default=0.75)
parser <- add_argument(parser, "--Bxy", help='coefficient of x on y', default=0.3)
parser <- add_argument(parser, "--Bzy", help='coeffficient of z on y', default=-0.3)
parser <- add_argument(parser, "--outcome_formula", help='formula for the outcome variable', default="y~x+z")
parser <- add_argument(parser, "--proxy_formula", help='formula for the proxy variable', default="w_pred~y")

args <- parse_args(parser)

B0 <- 0
Bxy <- args$Bxy
Bzy <- args$Bzy


if(args$m < args$N){
    df <- simulate_data(args$N, args$m, B0, Bxy, Bzy, args$seed, args$prediction_accuracy)

#    result <- list('N'=args$N,'m'=args$m,'B0'=B0,'Bxy'=Bxy,'Bzy'=Bzy, 'seed'=args$seed, 'y_explained_variance'=args$y_explained_variance, 'prediction_accuracy'=args$prediction_accuracy, 'x_bias_y0'=args$x_bias_y0,'x_bias_y1'=args$x_bias_y1,'outcome_formula' = args$outcome_formula, 'proxy_formula' = args$proxy_formula)
    result <- list('N'=args$N,'m'=args$m,'B0'=B0,'Bxy'=Bxy,'Bzy'=Bzy, 'seed'=args$seed, 'y_explained_variance'=args$y_explained_variance, 'prediction_accuracy'=args$prediction_accuracy, 'outcome_formula' = args$outcome_formula, 'proxy_formula' = args$proxy_formula)

    outline <- run_simulation_depvar(df, result, outcome_formula = as.formula(args$outcome_formula), proxy_formula = as.formula(args$proxy_formula))

    outfile_lock <- lock(paste0(args$outfile, '_lock'),exclusive=TRUE)

    if(file.exists(args$outfile)){
        logdata <- read_feather(args$outfile)
        logdata <- rbind(logdata,as.data.table(outline),fill=TRUE)
    } else {
        logdata <- as.data.table(outline)
    }

    print(outline)
    write_feather(logdata, args$outfile)
    unlock(outfile_lock)
}
